#include <qdatetime.h>
#include <qfiledialog.h>
#include <qimage.h>
#include <qmessagebox.h>
#include <qpainter.h>
#include <qsettings.h>
#include <qstatusbar.h>
#include <qthread.h>
#include <qtimer.h>

#include <math.h>

// required by frmSettings
#include <qcheckbox.h>
#include <qradiobutton.h>
#include <qspinbox.h>

#include "frmsettings.h"

// required by frmQuality
#include <qslider.h>

#include "frmquality.h"

const QString appName = "linman";
const QString appVer = "1.1";
const QString appAuthor = "amdfanatyk";
const QString appAuthorEmail = "amdfanatyk@o2.pl";
const QString appModifier = ""; // if You have made some changes, type Your name/nick here
const QString appModifierEmail = ""; // if You have made some changes, type Your email here

bool
	var = 1, // variable iteration depth?
	mouseDown = false, // mouse button pressed?
	forceRedraw = false, // force image redraw?
	frame, // draw frame?
	modified; // image state
int
	iter = 0,
	color = 2, // 0 - black&white, 1 - percent, 2 - iteration
	rFactor = 5, // red factor
	gFactor = 10, // green factor
	bFactor = 15, // blue factor
	zoom, // zoom ;)
	clientSizeX, // width of the window
	clientSizeY, // height of the window - 60
	oldClientSizeX = -1, // previous width of the window
	oldClientSizeY = -1; // prevoius height of the window	
struct complex
{
	double x;
	double y;
}
	min,
	prec,
	range,
	mouseDownOn;
QImage
	image; // fractal
QPoint
	mousePressedOn;
QSettings
	settings;
QStatusBar
	* statBar = NULL;
QString
	currentFileName = "",
	currentDirName = "";
QTimer
	timer;

class CDraw : public QThread
{
	public:
		CDraw(const int & sX, const int & sY, const int & hX, const int & hY, const complex & wsp, const double & os)
		{
			startX = sX;
			startY = sY;
			stopX = hX;
			stopY = hY;
			c = wsp;
			offset = os;
		}

		~CDraw()
		{
		}

		void run()
		{
			complex z, old_z;

			for (int y = startY ; y < stopY ; y++)
			{
				for (int x = startX ; x < stopX ; x++)
				{
					z.x = 0;
					z.y = 0;

					old_z.x = 0;
					old_z.y = 0;

					unsigned int k = 100;

					if (var)
					{
						if (range.x <= 0.1)
							k = 1000;

						if (range.x <= 0.01)
							k = 10000;
				
						if (range.x <= 0.003)
							k = 33333;
					}

					unsigned int tmp_iter = k;

					do
					{
						z.x = old_z.x * old_z.x - old_z.y * old_z.y + c.x;
						z.y = 2 * old_z.x * old_z.y + c.y;
						old_z.x = z.x;
						old_z.y = z.y;
						k--;
					}
					while (k > 0 && z.x * z.x + z.y * z.y <= 4);

					iter += (tmp_iter - k);

					switch (color)
					{
						case 0:
							if (k == 0)
								image.setPixel(x, y, qRgb(0, 0, 0));
							else
								image.setPixel(x, y, qRgb(255, 255, 255));
						break;
						case 1:
							if (k == 0)
								image.setPixel(x, y, qRgb(0, 0, 0));
							else
							{
								int color_percent = qRgb(255, 0, 0);

								if ( (tmp_iter - k + 1) * 100 / tmp_iter <= 20)
									color_percent = qRgb(0, 255, 0);

								if ( (tmp_iter - k + 1) * 100 / tmp_iter <= 10)
									color_percent = qRgb(0, 0, 255);

								image.setPixel(x, y, color_percent);
							}
						break;
						case 2:
							image.setPixel(x, y,
								qRgb(rFactor * k, gFactor * k, bFactor * k));
						break;
					}
			
					c.x += prec.x;
				}

				c.y += prec.y;
				c.x = min.x + offset;
			}
		}

	private:
		int startX;
		int startY;
		int stopX;
		int stopY;
		complex c;
		double offset;
};

void frmMain::init()
{
	min.x = -2.3;
	min.y = -1.2;
	range.x = 3.2;
	range.y = 2.4;
	
	connect(&timer, SIGNAL(timeout()), this, SLOT(onTimer()));

	statBar = statusBar();
	
	settings.setPath("amdfanatyk", "Software");

	readData();

	onModified(false);

	timer.start(0, true);
}

void frmMain::destroy()
{
	writeData();
}

void frmMain::onTimer()
{
	forceRedraw = true;

	onDraw();
}

void frmMain::paintEvent(QPaintEvent * e)
{
	Q_UNUSED(e);

	onDraw();
}

void frmMain::resizeEvent(QResizeEvent * e)
{
	Q_UNUSED(e);

	clientSizeX = width();
	clientSizeY = (height() - 60);

	image.create(clientSizeX, clientSizeY, 32);
}

void frmMain::mousePressEvent(QMouseEvent * e)
{
	mouseDownOn.x = e->x() * range.x / clientSizeX;
	mouseDownOn.y = (e->y() - 30) * range.y / clientSizeY;
		
	mousePressedOn.setX(e->x());
	mousePressedOn.setY(e->y());
	
	switch ((int)e->stateAfter())
	{
		case Qt::LeftButton:
			mouseDown = true;
			frame = true;
		break;
		case (Qt::LeftButton | Qt::ShiftButton):
			frame = false;
		break;
        }

	e->accept();
}

void frmMain::mouseMoveEvent(QMouseEvent * e)
{
	if (mouseDown && frame)
	{
		QPainter painter(this);
		painter.setPen(QPen(white));
		painter.setRasterOp(XorROP);
		if (e->x() - mousePressedOn.x() >= 0 && e->y() - mousePressedOn.y() >= 0)
		{
			painter.drawRect(mousePressedOn.x(), 
				mousePressedOn.y(),
					e->x() - mousePressedOn.x(),
						e->y() - mousePressedOn.y());
			repaint(mousePressedOn.x(), 
					mousePressedOn.y(), 
						e->x() - 2 * mousePressedOn.x() + width(),
							 e->y() - 2 * mousePressedOn.y() + height(),
								false);
			painter.drawRect(mousePressedOn.x(), 
				mousePressedOn.y(),
					e->x() - mousePressedOn.x(),
						e->y() - mousePressedOn.y());
		}
	}
}

void frmMain::mouseReleaseEvent(QMouseEvent * e)
{
	switch ((int)e->state())
	{
		case Qt::LeftButton:
			min.x += mouseDownOn.x;
			min.y += mouseDownOn.y;
			range.x = fabs(e->x() * range.x / clientSizeX - mouseDownOn.x);
			range.y = 2.4 / 3.2 * range.x;
		break;
		case (Qt::LeftButton | Qt::ShiftButton):
			min.x += e->x() * range.x / clientSizeX - mouseDownOn.x;
			min.y += (e->y() - 30) * range.y / clientSizeY - mouseDownOn.y;
		break;
		case Qt::RightButton:
			if (range.x * 2 < 3.2)
			{
				min.x -= range.x / 2;
				min.y -= range.y / 2;
				range.x *= 2;
				range.y *= 2;	
			}
			else
			{
				min.x = -2.3;
				min.y = -1.2;
				range.x = 3.2;
				range.y = 2.4;
			}
		break;
		case (Qt::RightButton | Qt::ControlButton):
			min.x = -2.3;
			min.y = -1.2;
			range.x = 3.2;
			range.y = 2.4;
		break;
	}

	mouseDown = false;
	forceRedraw = true;

	onModified(true);

	onDraw();
}

void frmMain::onDraw()
{
	iter = 0; // number of iterations
	
	// calculates precision
	prec.x = range.x / clientSizeX;
	prec.y = range.y / clientSizeY;

	zoom = (unsigned int)(3.2 / range.x);

	QPainter painter(this);

	if (clientSizeX != oldClientSizeX || clientSizeY != oldClientSizeY || forceRedraw)
	{
		statBar->message("Drawing in progress, please wait...");
		setCursor(Qt::WaitCursor);
		QTime time;
		time.start();

		int hClientSizeX = clientSizeX / 2;
		int hClientSizeY = clientSizeY / 2;

		complex c1 = {min.x + hClientSizeX * prec.x, min.y};
		complex c2 = {min.x, min.y};
		complex c3 = {min.x, min.y + hClientSizeY * prec.y};
		complex c4 = {min.x + hClientSizeX * prec.x, min.y + hClientSizeY * prec.y};

		CDraw
			draw1(hClientSizeX, 0, clientSizeX, hClientSizeY, c1, hClientSizeX * prec.x),
			draw2(0, 0, hClientSizeX, hClientSizeY, c2, 0),
			draw3(0, hClientSizeY, hClientSizeX, clientSizeY, c3, 0),
			draw4(hClientSizeX, hClientSizeY, clientSizeX, clientSizeY, c4, hClientSizeX * prec.x);

		draw1.start(QThread::TimeCriticalPriority);
		draw2.start(QThread::TimeCriticalPriority);
		draw3.start(QThread::TimeCriticalPriority);
		draw4.start(QThread::TimeCriticalPriority);
		draw1.wait();
		draw2.wait();
		draw3.wait();
		draw4.wait();

		oldClientSizeX = clientSizeX;
		oldClientSizeY = clientSizeY;
		
		forceRedraw = false;

		onModified(modified);

		unsigned int elapsed = time.elapsed();
		unsigned int speed = 0;

		if (elapsed > 0)
			speed = (unsigned int)(iter / elapsed);
				
		setCursor(Qt::ArrowCursor);

		statBar->message(QString::number(iter) + " iterations done in " + QString::number(elapsed) +
			" ms (" + QString::number(speed) + " iter/ms)");
	}
	
	painter.drawImage(0, 30, image);
}

void frmMain::writeData()
{
	settings.writeEntry("/linman/fractal/min_x", min.x);
	settings.writeEntry("/linman/fractal/min_y", min.y);
	settings.writeEntry("/linman/fractal/range_x", range.x);
	settings.writeEntry("/linman/fractal/range_y", range.y);
	settings.writeEntry("/linman/fractal/var", var);
	settings.writeEntry("/linman/fractal/color", color);
	settings.writeEntry("/linman/fractal/rFactor", rFactor);
	settings.writeEntry("/linman/fractal/gFactor", gFactor);
	settings.writeEntry("/linman/fractal/bFactor", bFactor);
	settings.writeEntry("/linman/geometry/x", geometry().x());
	settings.writeEntry("/linman/geometry/y", geometry().y());
	settings.writeEntry("/linman/geometry/width", geometry().width());
	settings.writeEntry("/linman/geometry/height", geometry().height());

	onModified(false);
	
	statBar->message("Settings saved.");
}

void frmMain::readData()
{
	min.x = settings.readDoubleEntry("/linman/fractal/min_x", min.x);
	min.y = settings.readDoubleEntry("/linman/fractal/min_y", min.y);
	range.x = settings.readDoubleEntry("/linman/fractal/range_x", range.x);
	range.y = settings.readDoubleEntry("/linman/fractal/range_y", range.y);
	var = settings.readBoolEntry("/linman/fractal/var", var);
	color = settings.readNumEntry("/linman/fractal/color", color);
	rFactor = settings.readNumEntry("/linman/fractal/rFactor", rFactor);
	gFactor = settings.readNumEntry("/linman/fractal/gFactor", gFactor);
	bFactor = settings.readNumEntry("/linman/fractal/bFactor", bFactor);
	setGeometry
	(
		settings.readNumEntry("/linman/geometry/x", geometry().x()),
		settings.readNumEntry("/linman/geometry/y", geometry().y()),
		settings.readNumEntry("/linman/geometry/width", geometry().width()),
		settings.readNumEntry("/linman/geometry/height", geometry().height())
	);

	forceRedraw = true;
	onModified(false);

	statBar->message("Settings restored.");

	onDraw();
}

void frmMain::writeData(QString fileName)
{
	QDir dir;
	QFile fout;
	QTextStream streamout;

	if (fileName.isEmpty())
		return;

	fout.setName(fileName);
	streamout.setDevice(&fout);
	fout.open(IO_WriteOnly);
		streamout << 
			QString::number(min.x) << "\n" <<
				QString::number(min.y) << "\n" <<
					QString::number(range.x) << "\n" <<
						QString::number(range.y) << "\n" <<
							QString::number(var) << "\n" <<
								QString::number(color) << "\n" <<
									QString::number(rFactor) << "\n" <<
										QString::number(gFactor) << "\n" <<
											QString::number(bFactor) << "\n";
	fout.close();

	onModified(false);
	
	statBar->message("Saved to file successfully.");
}

void frmMain::readData(QString fileName)
{
	QDir dir;
	QFile fin;
	QTextStream streamin;

	if (fileName.isEmpty())
	{
		statBar->message("Read from file failed.");
		return;
	}
	
	fin.setName(fileName);

	fin.setName(fileName);
	streamin.setDevice(&fin);
	fin.open(IO_ReadOnly);
		min.x = streamin.readLine().toDouble();
		min.y = streamin.readLine().toDouble();
		range.x = streamin.readLine().toDouble();
		range.y = streamin.readLine().toDouble();
		var = streamin.readLine().toInt();
		color = streamin.readLine().toInt();
		rFactor = streamin.readLine().toInt();
		gFactor = streamin.readLine().toInt();
		bFactor = streamin.readLine().toInt();
	fin.close();
		
	forceRedraw = true;
	onModified(false);
		
	statBar->message("Read from file successfully.");

	onDraw();
}

void frmMain::onClose()
{
	int res = -1;

	if (modified)
	{
		QMessageBox msg("Question", "Do You want to save changes to file?", QMessageBox::Warning, QMessageBox::Yes,
			QMessageBox::No, QMessageBox::Cancel);

		res = msg.exec();
	
		if (res == QMessageBox::Yes)
			onSave();
	}	
	
	if (res != QMessageBox::Cancel)
	{
		var = 1;
		color = 2;
		
		rFactor = 5;
		gFactor = 10;
		bFactor = 15;	

		min.x = -2.3;
		min.y = -1.2;
		range.x = 3.2;
		range.y = 2.4;

		forceRedraw = true;
		currentFileName = "";

		onModified(false);
		
		onDraw();
	}
}

void frmMain::onExport()
{
	if (currentDirName.isEmpty())
		currentDirName = QDir::homeDirPath();

	QString fileName = QFileDialog::getSaveFileName(currentDirName, "png files (*.png)", this, "", "Export");
	
	if (!fileName.isEmpty())	
	{
		frmQuality frm(this);
		frm.exec();
		
		currentDirName = QFileInfo(currentFileName).dirPath(true);

		if (fileName.right(4) != ".png" )
		{
			if (fileName.right(4) != ".PNG")
				fileName += ".png";
		}
		
		bool res = image.save(fileName, "PNG", frm.sldQuality->value());
	
		if (res)
			statBar->message("Exported to file successfully.");
		else
			statBar->message("Exporting to file failed.");	
	}
}

void frmMain::onSettings()
{
	frmSettings frm(this);
	
	frm.spnRed->setValue(rFactor);
	frm.spnGreen->setValue(gFactor);
	frm.spnBlue->setValue(bFactor);

	if (var)
		frm.rdoHigh->setChecked(1);
	else
		frm.rdoLow->setChecked(1);

	switch (color)
	{
		case 0:
			frm.rdoBlackWhite->setChecked(1);
		break;
		case 1:
			frm.rdoPercent->setChecked(1);
		break;
		case 2:
			frm.rdoIteration->setChecked(1);
			frm.spnRed->setEnabled(1);
			frm.spnGreen->setEnabled(1);
			frm.spnBlue->setEnabled(1);
		break;
	}
	
	frm.exec();
	
	bool newVar = frm.rdoHigh->isChecked();
	
	int newColor;

	if (frm.rdoBlackWhite->isChecked())
		newColor = 0;
	else
	{	
		if (frm.rdoPercent->isChecked())
			newColor = 1;
		else
			newColor = 2;
	}

	int newRFactor = frm.spnRed->value();
	int newGFactor = frm.spnGreen->value();
	int newBFactor = frm.spnBlue->value();

	if (newVar != var || newColor != color || newRFactor != rFactor || newGFactor != gFactor || newBFactor != bFactor)
	{
		var = newVar;
		color = newColor;

		rFactor = newRFactor;
		gFactor = newGFactor;
		bFactor = newBFactor;

		forceRedraw = true;
		onModified(true);

		onDraw();
	}
}

void frmMain::onOpen()
{
	QString storCurrentFileName(currentFileName), storCurrentDirName(currentDirName);

	if (modified)
	{
		QMessageBox msg("Question", "Do You want to save changes to file?", QMessageBox::Warning, QMessageBox::Yes,
			QMessageBox::No, QMessageBox::NoButton);
		
		if (msg.exec() == QMessageBox::Yes)
			onSave();
	}
	
	currentFileName = storCurrentFileName;
	currentDirName = storCurrentDirName;

	if (currentDirName.isEmpty())
		currentDirName = QDir::homeDirPath();

	currentFileName = QFileDialog::getOpenFileName(currentDirName, "linman files (*.lm)", this, "", "Open");
	
	if (!currentFileName.isEmpty())	
	{
		currentDirName = QFileInfo(currentFileName).dirPath(true);
	
		readData(currentFileName);
	}
}

void frmMain::onSave()
{
	if (!currentFileName.isEmpty() && modified)
		writeData(currentFileName);
	else
	{
		if (modified)
			onSaveAs();
	}
}

void frmMain::onSaveAs()
{
	QString fileName;

	if (currentDirName.isEmpty())
		currentDirName = QDir::homeDirPath();

	fileName = QFileDialog::getSaveFileName(currentDirName, "linman files (*.lm)", this, "", "Save as");
	
	if (!fileName.isEmpty())	
	{
		QFile fout(fileName);

		currentDirName = QFileInfo(fileName).dirPath(true);

		if (fileName.right(3) != ".lm")
			fileName += ".lm";

		int res = -1;

		if (fout.exists())
		{
			QMessageBox msg("Question", "File already exists!\nDo You want to overwrite it?", QMessageBox::Warning,
				QMessageBox::Yes, QMessageBox::No, QMessageBox::NoButton);
			res = msg.exec();
		}

		if (res != QMessageBox::No)
		{
			currentFileName = fileName;
			writeData(currentFileName);
		}
	}
}

void frmMain::onAbout()
{
	QString message("<b>" + appName + "</b>-<i>" + appVer + "</i><br>licence: <b>GPLv2</b><br><br><b>programming:</b><br>" + appAuthor + "<br><br><b>email:</b><br>" + appAuthorEmail);

	if (!appModifier.isEmpty() && !appModifierEmail.isEmpty())
		message += ("<br><br><b>modified by:</b><br>" + appModifier + "<br><br><b>email:</b><br>" + appModifierEmail);

	QMessageBox msg("About", message,
		QMessageBox::Information, QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
	msg.exec();
}

void frmMain::onAboutQt()
{
	QMessageBox::aboutQt(this);
}

void frmMain::closeEvent( QCloseEvent * e )
{
	int res = -1;

	if (modified)
	{
		QMessageBox msg("Question", "Do You want to save changes to file?", QMessageBox::Warning, QMessageBox::Yes,
			QMessageBox::No, QMessageBox::Cancel);

		res = msg.exec();
	
		if (res == QMessageBox::Yes)
			onSave();
	}
	
	if (res != QMessageBox::Cancel)
	{
		writeData("");

		close();

		e->accept();
	}
	else
		e->ignore();
}

void frmMain::onExit()
{
	close();
}

void frmMain::onModified(bool newValue)
{
	modified = newValue;

	if (modified)
		setCaption("linman " + QString::number(zoom) + "x" + " [" + currentFileName + "*]");
	else
 		setCaption("linman " + QString::number(zoom) + "x" + " [" + currentFileName + "]");
}

void frmMain::onKeys()
{
	QMessageBox msg("Keys",
		"button\t\t-\tusage\t\t-\teffect\n\nLMB\t\t-\tdrag & drop\t-\tzoom in\nLMB + Shift\t-\tdrag & drop\t-\tmove\nRMB\t\t-\tclick\t\t-\tzoom out\nRMB + Ctrl\t-\tclick\t\t-\tdefault size",
			QMessageBox::Information, QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
	msg.exec();
}
